﻿unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants,
  System.Classes, Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs,
  VclTee.TeeGDIPlus, VCLTee.TeEngine, Vcl.ExtCtrls, VCLTee.TeeProcs,
  VCLTee.Chart, Vcl.StdCtrls, VCLTee.Series, VCLTee.TeeTools,
  Generics.Collections;

type
  TCursorInfoRecord = record
    StrDate: String;
    StrTime: String;
    YValue: Double;
  end;

  TForm1 = class(TForm)
    Chart1: TChart;
    Panel1: TPanel;
    Button1: TButton;
    Button2: TButton;
    Series1: TLineSeries;
    LabeledEdit1: TLabeledEdit;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Chart1GetAxisLabel(Sender: TChartAxis; Series: TChartSeries;
      ValueIndex: Integer; var LabelText: string);
    procedure Chart1AfterDraw(Sender: TObject);
    procedure CheckBox1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Chart1Zoom(Sender: TObject);
    procedure Chart1UndoZoom(Sender: TObject);
  private
    FDisplayRange: Integer;
    FInitialized: Boolean;
    FCurrentCursorValues: TCursorInfoRecord;
    FStartDate: TDateTime;
    FValues: TList<Double>;
    FZoomData: TList<Cardinal>;

    procedure ClearZoomData(AChart: TChart; ARepaint: Boolean);
    procedure InsertZoomData(AChart: TChart; AFromX, AToX: Cardinal);

    procedure InitializeObserverChart(var AChart: TChart);
    procedure CheckChartBounds(AChart: TChart; AValue: Extended;
      AIncrement: Integer = 100);
    procedure SetChartSeriesMinMax(AChart: TChart; AValue: Extended);
    procedure HandleScroll(AChart: TChart);
    procedure ChartTool_CursorChange(Sender: TCursorTool; x, y: Integer;
      const XValue, YValue: Double; Series: TChartSeries; ValueIndex: Integer);
    function GetXValueIndex(ASeries: TCustomLineSeries; X: Integer): Integer;
  public
    { Public-Deklarationen }
  end;

var
  Form1: TForm1;

implementation

uses
  Math, DateUtils, VCLTee.TeeThemes;

{$R *.dfm}

procedure TForm1.Button1Click(Sender: TObject);
begin
  Randomize;
  FValues.Clear;
  FDisplayRange := -1;
  InitializeObserverChart(Chart1);
  SetChartSeriesMinMax(Chart1, 10);
  FStartDate := Now;
  FInitialized := TRUE;
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  SWR, R, NR, DR: Double;
  N: TDateTime;
  i, b, Nth: Integer;
  MaxValues: Cardinal;
  S: String;
begin
  if not FInitialized then
    Button1.Click;

  Chart1.BottomAxis.LabelStyle := talPointValue;
  FCurrentCursorValues.StrDate := '';
  FCurrentCursorValues.StrTime := '';
  FCurrentCursorValues.YValue := -1;

  FValues.Clear;

  MaxValues := String(LabeledEdit1.Text).ToInteger;
  b := 1;
  R := 0;
  for i := 1 to MaxValues do
    begin
      NR := Random * b;
      R := R + (b - R) * NR;
      DR := Sin(i * (Pi / 180)) * (R * 10);
      FValues.Add(DR);
      if (Random(200) < 50) or (Abs(DR) > 750) then
        b := -b;
    end;

  N := FStartDate;
  Nth := Round(Max(FValues.Count / Chart1.ClientWidth, 1));
  for i := 0 to FValues.Count - 1 do
    begin
      if (i mod Nth = 0) or (i = FValues.Count - 1) then
        begin
          N := N + i / (24 * 60 * 60);
          S :=
            FormatDateTime('ddddd', N) + #13#10 +
            FormatDateTime('tt', N);
          DR := FValues[i];
          Chart1.Series[0].AddXY(i, DR, S, clBlue);
          CheckChartBounds(Chart1, DR);
        end;
    end;
  HandleScroll(Chart1);
end;

procedure TForm1.HandleScroll(AChart: TChart);
var
  rFrom, rTo, rRange, rLen: Integer;
begin
  with AChart.Series[0] do
    rLen := Ceil(XValue[Count - 1]);
  rRange := 100;
  if FDisplayRange = -1 then
    begin
      rFrom := 0;
      rTo := rLen + rRange;
    end
  else
    begin
      rFrom := Max(rLen - FDisplayRange, 0);
      rTo := rLen + 10;
    end;
  if rFrom < rTo then
  	begin
      AChart.Axes.Bottom.Maximum := rTo;
      AChart.Axes.Bottom.Minimum := rFrom;
    end;
end;

procedure TForm1.SetChartSeriesMinMax(AChart: TChart;
  AValue: Extended);
begin
  AChart.LeftAxis.AutomaticMaximum := FALSE;
	AChart.LeftAxis.AutomaticMinimum := FALSE;
	AChart.LeftAxis.Minimum := -AValue;
	AChart.LeftAxis.Maximum := AValue;
end;

procedure TForm1.Chart1AfterDraw(Sender: TObject);
var
  Chart: TChart;
  mp: TPoint;
  R: TRect;
  DT: TDateTime;
  X, Y: Integer;
  sz1, sz2, sz: TSize;
  S1, S2: String;
begin
  // Draw center line
  with TChart(Sender), Canvas do
    begin
      R := ChartRect;
      Y := Axes.Left.CalcPosValue(0);
      Pen.Mode := pmCopy;;
      Pen.Color := clBlack;
      Pen.Width := 1;
      MoveTo(R.Left, Y);
      LineTo(R.Right, Y);
    end;

  // Draw cursor information, if available
  if ((FCurrentCursorValues.StrDate.Length = 0) and
      (FCurrentCursorValues.StrTime.Length = 0) and
      (FCurrentCursorValues.YValue = -1)) or
      (not (Sender is TChart)) then
    Exit;

  Chart := TChart(Sender);
  mp := Chart.ScreenToClient(Mouse.CursorPos);

  X := mp.X + 20;
  Y := mp.Y + 20;
  with Chart.Canvas do
    begin
      S1 := FCurrentCursorValues.StrDate;
      sz1.cx := TextWidth(S1);
      sz1.cy := TextHeight(S1);

      S2 := FCurrentCursorValues.StrTime;
      sz2.cx := TextWidth(S2);
      sz2.cy := TextHeight(S2);

      if (sz1.cx > sz2.cx) then
        sz.cx := sz1.cx
      else
        sz.cx := sz2.cx;
      sz.cx := sz.cx + 8;
      sz.cy := sz1.cy + sz2.cy + 8;

      R := Rect(X, Y, X + sz.cx, Y + sz.cy);

      Brush.Color := clInfoBk;
      Pen.Color := clBlack;
      Rectangle(R);

      TextOut(R.Left + 4, R.Top + 4, S1);
      TextOut(R.Left + 4, R.Top + sz1.cy + 4, S2);
    end;

  Finalize(FCurrentCursorValues.StrDate);
  Finalize(FCurrentCursorValues.StrTime);
end;

procedure TForm1.Chart1GetAxisLabel(Sender: TChartAxis; Series: TChartSeries;
  ValueIndex: Integer; var LabelText: string);
var
	S, StrDate, StrTime: String;
  DT: TDateTime;
  FS: TFormatSettings;
begin
	if (Sender <> Chart1.Axes.Bottom) or
      (ValueIndex < 0) or
      (Series.Tag <> $FF) then
  	Exit;

  LabelText := Series.XLabel[ValueIndex];
end;

procedure TForm1.Chart1UndoZoom(Sender: TObject);
begin
  ClearZoomData(TChart(Sender), TRUE);
end;

procedure TForm1.Chart1Zoom(Sender: TObject);
var
  Chart: TChart;
begin
  Chart := TChart(Sender);
  InsertZoomData(Chart, Chart.Zoom.X0, Chart.Zoom.X1);
end;

procedure TForm1.ChartTool_CursorChange(Sender: TCursorTool; x,
  y: Integer; const XValue, YValue: Double; Series: TChartSeries;
  ValueIndex: Integer);
var
  i: Integer;
  S: String;
begin
  if not (Series is TLineSeries) then
    Exit;

  // We need the series index of the X value
  i := GetXValueIndex(TLineSeries(Series), X);
  if i < 0 then
    Exit;

  S := Series.Labels[i];
  SetLength(FCurrentCursorValues.StrDate, 10);
  SetLength(FCurrentCursorValues.StrTime, 8);

  Move(S[1], FCurrentCursorValues.StrDate[1], 10 * SizeOf(Char));
  Move(S[13], FCurrentCursorValues.StrTime[1], 8 * SizeOf(Char));

  FCurrentCursorValues.YValue := YValue;
  Sender.ParentChart.Draw;
end;

procedure TForm1.CheckBox1Click(Sender: TObject);
begin
  Button1.Click;
end;

procedure TForm1.CheckChartBounds(AChart: TChart;
  AValue: Extended; AIncrement: Integer);
begin
  if (AValue > AChart.LeftAxis.Maximum) or (AValue < AChart.LeftAxis.Minimum) or
  	(-AValue > AChart.LeftAxis.Maximum) or (-AValue < AChart.LeftAxis.Minimum)
    then
      SetChartSeriesMinMax(AChart, Abs(AValue) + AIncrement);
end;

procedure TForm1.ClearZoomData(AChart: TChart; ARepaint: Boolean);
var
  i, j: Integer;
begin
  AChart.AutoRepaint := FALSE;
  AChart.Series[0].Active := FALSE;
  try
    for i := AChart.Series[0].Count - 1 downto 0 do
      for j := FZoomData.Count - 1 downto 0 do
        if Trunc(AChart.Series[0].XValue[i]) = FZoomData[j] then
          begin
            AChart.Series[0].Delete(i);
            FZoomData.Delete(j);
          end;
    FZoomData.Clear;
  finally
    AChart.Series[0].Active := TRUE;
    AChart.AutoRepaint := TRUE;
    if ARepaint then
      AChart.Refresh;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  ReportMemoryLeaksOnShutdown := TRUE;
  FValues := TList<Double>.Create;
  FZoomData := TList<Cardinal>.Create;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FValues.Free;
  FZoomData.Free;
end;

function TForm1.GetXValueIndex(ASeries: TCustomLineSeries;
  X: Integer): Integer;
var
  i: Integer;
  Value: Double;
begin
  Value := ASeries.ParentChart.BottomAxis.CalcPosPoint(X);
  for i := ASeries.FirstValueIndex to ASeries.LastValueIndex do
    begin
      if (i >= 0) and (ASeries.XValue[i] > Value) then
        begin
          Result := i;
          Exit;
        end;
    end;

  Result := ASeries.LastValueIndex;
end;

procedure TForm1.InitializeObserverChart(var AChart: TChart);
var
  Ls: TLineSeries;
  i: Integer;
  CursorTool: TCursorTool;
begin
  with AChart do
    begin
      Tools.Clear;
      SeriesList.Clear;

      Hover.Visible := FALSE;

      MarginLeft := 1;
      MarginRight := 1;
      MarginBottom := 2;
      MarginTop := 2;

      View3D := FALSE;
      Color := clWhite;
      Legend.Visible := FALSE;

      LeftAxis.Title.Caption := 'Z';

      BottomAxis.LabelStyle := talText;
      BottomAxis.LabelsMultiLine := TRUE;

      Title.Text.Clear;
      Title.Text.Add(FormatDateTime('ddddd', Now));

      Ls := TLineSeries.Create(AChart);
      Ls.Title := 'Z';
      Ls.ParentChart := AChart;
      Ls.XValues.DateTime := FALSE;
      Ls.Pen.Width := 1;
      Ls.Color := clGreen;
      Ls.Tag := $FF;
      Ls.XValues.Order := loAscending;

      with Axes.Left, Grid do
        begin
          Grid.Color := RGB(222, 222, 222);
          SmallDots := TRUE;
          GridCentered := TRUE;
          AutomaticMaximum := FALSE;
          AutomaticMinimum := FALSE;
          Maximum := 100;
          Minimum := -Maximum;
        end;

      with Axes.Bottom, Grid do
        begin
          Grid.Color := RGB(222, 222, 222);
          SmallDots := TRUE;
          GridCentered := TRUE;
          AutomaticMaximum := FALSE;
          AutomaticMinimum := FALSE;
          Minimum := 0;
          Maximum := 1;
        end;
    end;

  CursorTool := TCursorTool.Create(AChart);
  with CursorTool do
    begin
      ParentChart := AChart;
      Series := AChart.Series[0];
      Style := cssScope;
      ScopeSize := 4;
      ScopeStyle := scsRectangle;
      FollowMouse := TRUE;
      AxisAnnotation.Visible := FALSE;
      AxisAnnotation.Active := FALSE;
      OnChange := ChartTool_CursorChange;
    end;
end;

procedure TForm1.InsertZoomData(AChart: TChart; AFromX, AToX: Cardinal);
var
  i, IndexFrom, IndexTo: Cardinal;
  S: String;
  Nth, IndexLen: INteger;
  N: TDateTime;
  DR: Double;
begin
  ClearZoomData(AChart, FALSE);

  IndexFrom := Floor(AChart.BottomAxis.CalcPosPoint(AFromX));
  IndexTo := Ceil(AChart.BottomAxis.CalcPosPoint(AToX));
  IndexLen := IndexTo - IndexFrom;
  if IndexLen <= 0 then
    Exit;
  Nth := Ceil(IndexLen / Max(AChart.ClientWidth, 1));

  AChart.AutoRepaint := FALSE;
  AChart.Series[0].BeginUpdate;
  try
    N := FStartDate;
    for i := IndexFrom to IndexTo do
      begin
        if (i mod Nth = 0) or (i = IndexTo) then
          begin
            FZoomData.Add(i);
            N := N + i / (24 * 60 * 60);
            S :=
              FormatDateTime('ddddd', N) + #13#10 +
              FormatDateTime('tt', N);
            DR := FValues[i];
            Chart1.Series[0].AddXY(i, DR, S, clGreen);
          end;
      end;
  finally
    AChart.Series[0].EndUpdate;
    AChart.AutoRepaint := TRUE;
    AChart.Refresh;
  end;
end;

end.
